import java.awt.LayoutManager;
import java.awt.LayoutManager2;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;

/**
 * Layout similar to BoxLayout and FlowLayout, but with more powerful features.
 * <p>
 * Components are arranged along the X_AXIS or the Y_AXIS
 * in the order that they were added to their container.
 * <p>
 * The <i>alignment</i> determines how components are aligned
 * relative to each other. An alignment of LEFT, for example,
 * aligns the left edges of the components.
 * The special alignment FILL grows all components to either
 * 1) the height of the container (if X_AXIS is specified) or
 * 2) the width of the container (if Y_AXIS is specified)
 * <p>
 * This layout supports <tt>Boolean</tt> constraints. If a component
 * is added with the constraint <tt>Boolean.TRUE</tt>, it will be designated
 * as the <i>extra space absorber</i>. Such a component will be grown to fill
 * any gap that occurs along the axis of the layout. By default, all components
 * added to the layout will have the constraint <tt>Boolean.FALSE</tt>.
 *
 * @author David Foster <davidfstr@gmail.com>
 */
public class ListLayout implements LayoutManager, LayoutManager2 {
	public final static int X_AXIS	= 1;
	public final static int Y_AXIS	= 2;
	
	private final static int MIN	= 1;
	private final static int MAX	= 2;
	
	public final static int TOP		= MIN;
	public final static int BOTTOM	= MAX;
	public final static int LEFT	= MIN;
	public final static int RIGHT	= MAX;
	
	public final static int CENTER	= 3;
	public final static int FILL	= 4;
	
	private int axis;
	private int alignment;
	
	private Component extraSpaceAbsorber;
	
	private boolean valid;
	private GridBagLayout backingLayout;
	private GridBagConstraints backingConstraints = new GridBagConstraints();
	
	// ### INIT ###
	
	public ListLayout() {
		this(X_AXIS);
	}
	
	public ListLayout(int axis) {
		this(axis, FILL);
	}
	
	/**
	 * @param axis		axis along which components are laid;
	 * 					either X_AXIS or Y_AXIS.
	 * @param alignment	one of LEFT/RIGHT (axis=X_AXIS),
	 * 					TOP/BOTTOM (axis=Y_AXIS),
	 * 					CENTER, or FILL.
	 */
	public ListLayout(int axis, int alignment) {
		this.axis = axis;
		this.alignment = alignment;
		
		this.extraSpaceAbsorber = null;
		this.valid = false;
		this.backingLayout = new GridBagLayout();
	}
	
	// ### CONSTRAINTS ###
	
	public void addLayoutComponent(String constraints, Component comp) {
		this.addLayoutComponent(comp, constraints);
	}
	
	public void addLayoutComponent(Component comp, Object constraints) {
		boolean isAbsorber;
		if (constraints == null) {
			isAbsorber = false;
		} else if (constraints instanceof Boolean) {
			isAbsorber = ((Boolean) constraints).booleanValue();
		} else {
			throw new IllegalArgumentException(
				"expected constraints to be of type Boolean");
		}
		
		if (isAbsorber) {
			if (extraSpaceAbsorber != null) {
				if (extraSpaceAbsorber != comp) {
					throw new IllegalStateException(
						"only one component can absorb extra space "+
						"(have a constraint of Boolean.TRUE)");
				}
			} else {
				extraSpaceAbsorber = comp;
			}
		} else {
			if (extraSpaceAbsorber == comp) {
				extraSpaceAbsorber = null;
			}
		}
		
		this.invalidate();
	}
	
	public void removeLayoutComponent(Component comp) {
		if (extraSpaceAbsorber == comp)
			extraSpaceAbsorber = null;
		
		this.invalidate();
	}
	
	// ### LAYOUT ###
	
	public void layoutContainer(Container parent) {
		this.validate(parent);
		backingLayout.layoutContainer(parent);
	}
	
	public Dimension minimumLayoutSize(Container parent) {

		this.validate(parent);
		return backingLayout.minimumLayoutSize(parent);
	}
	
	public Dimension preferredLayoutSize(Container parent) {
		this.validate(parent);
		return backingLayout.preferredLayoutSize(parent);
	}
	
	public Dimension maximumLayoutSize(Container parent) {
		this.validate(parent);
		return backingLayout.maximumLayoutSize(parent);
	}
	
	public float getLayoutAlignmentX(Container parent) {
		this.validate(parent);
		return backingLayout.getLayoutAlignmentX(parent);
	}
	
	public float getLayoutAlignmentY(Container parent) {
		this.validate(parent);
		return backingLayout.getLayoutAlignmentY(parent);
	}
	
	public void invalidateLayout(Container parent) {
		if (this.valid)
			backingLayout.invalidateLayout(parent);
	}
	
	// ### LAYOUT INFO ###
	
	private void invalidate() {
		this.valid = false;
	}
	
	private void validate(Container parent) {
		if (this.valid)
			return;
		
		for (int i=0, n=parent.getComponentCount(); i<n; i++) {
			Component curComp = parent.getComponent(i);
			boolean curCompIsAbsorber = (extraSpaceAbsorber == curComp);
			
			backingConstraints.gridx = (axis == X_AXIS) ? i : 0;
			backingConstraints.gridy = (axis == X_AXIS) ? 0 : i;
			if (axis == X_AXIS) {
				if (alignment == TOP)
					backingConstraints.anchor = GridBagConstraints.NORTH;
				else if (alignment == BOTTOM)
					backingConstraints.anchor = GridBagConstraints.SOUTH;
				else if ((alignment == CENTER) || (alignment == FILL))
					backingConstraints.anchor = GridBagConstraints.CENTER;
				
				backingConstraints.weightx = curCompIsAbsorber ? 1.0 : 0.0;
				backingConstraints.weighty = 1.0;
				backingConstraints.fill = (alignment == FILL)
					? GridBagConstraints.BOTH
					: GridBagConstraints.HORIZONTAL;
				
			} else {
				if (alignment == LEFT)
					backingConstraints.anchor = GridBagConstraints.WEST;
				else if (alignment == RIGHT)
					backingConstraints.anchor = GridBagConstraints.EAST;
				else if ((alignment == CENTER) || (alignment == FILL))
					backingConstraints.anchor = GridBagConstraints.CENTER;
				
				backingConstraints.weightx = 1.0;
				backingConstraints.weighty = curCompIsAbsorber ? 1.0 : 0.0;
				backingConstraints.fill = (alignment == FILL)
					? GridBagConstraints.BOTH
					: GridBagConstraints.VERTICAL;
			}
			
			backingLayout.setConstraints(curComp, backingConstraints);
		}
		
		this.valid = true;
	}
}
